"""
University of Florida Thesis/Dissertation Template for Pyexlatex
"""
from plufthesis.doc_class_type import register_doc_type
# Causes doc class type to get registered in pyexlatex
register_doc_type()
from plufthesis.transformers import elevate_sections_by_one_level



from enum import Enum
from typing import Optional

from pyexlatex.models.item import SimpleItem


class DegreeType(SimpleItem):
    name = 'degreeType'

    def __init__(self, dtype: str):
        self.dtype = dtype
        super().__init__(self.name, dtype)


class DegreeYear(SimpleItem):
    name = 'degreeYear'

    def __init__(self, year: int):
        self.year = year
        super().__init__(self.name, str(year))


class DegreeMonth(SimpleItem):
    name = 'degreeMonth'

    def __init__(self, month: str):
        self.month = month
        super().__init__(self.name, month)


class Major(SimpleItem):
    name = 'major'

    def __init__(self, major: str):
        self.major = major
        super().__init__(self.name, major)


class ThesisTypes(str, Enum):
    DISSERTATION = 'Dissertation'
    THESIS = 'Thesis'


class ThesisType(SimpleItem):
    name = 'thesisType'

    def __init__(self, ttype: ThesisTypes):
        self.ttype = ttype
        super().__init__(self.name, ttype.value)


class Chair(SimpleItem):
    name = 'chair'

    def __init__(self, chair: str, co_chair: Optional[str] = None):
        self.chair = chair
        self.co_chair = co_chair
        modifiers: Optional[str] = None
        if co_chair:
            modifiers = self._wrap_with_bracket(co_chair)
        super().__init__(self.name, chair, pre_modifiers=modifiers)


import pyexlatex as pl
import pytest
from pyexlatex.logic.builder import build

from plufthesis.transformers import elevate_sections_by_one_level

SUB_SUB_SECTION = pl.SubSubSection("Sub sub content", title="Sub Sub section")
SUB_SECTION = pl.SubSection(["Sub content", SUB_SUB_SECTION], title="Sub section")
SECTION = pl.Section(["Section content", SUB_SECTION], title="Section")
CHAPTER = pl.Chapter(["Chapter content", SECTION], title='Chapter')

SECTIONS_BY_TEXT = r"""
\section{Section}
Section content
\subsection{Sub section}
Sub content
\subsubsection{Sub Sub section}
Sub sub content
"""

CHAPTERS_BY_TEXT = r"""
\chapter{Chapter}
Chapter content
""" + SECTIONS_BY_TEXT


def test_elevate_sections_objects_success():
    content = build([SECTION])
    revised = elevate_sections_by_one_level(content)
    assert (
        revised
        == "\\begin{chapter}{Section}\nSection content\n\\begin{section}{Sub section}\nSub content\n\\begin{subsection}{Sub Sub section}\nSub sub content\n\\end{subsection}\n\\end{section}\n\\end{chapter}"
    )


def test_elevate_sections_text_success():
    revised = elevate_sections_by_one_level(SECTIONS_BY_TEXT)
    assert (
        revised
        == "\n\\chapter{Section}\nSection content\n\\section{Sub section}\nSub content\n\\subsection{Sub Sub section}\nSub sub content\n"
    )


def test_elevate_sections_objects_chapter():
    content = build([CHAPTER])
    with pytest.raises(ValueError) as exc_info:
        elevate_sections_by_one_level(content)
    assert 'cannot convert' in str(exc_info.value)


def test_elevate_sections_text_chapter():
    with pytest.raises(ValueError) as exc_info:
        elevate_sections_by_one_level(CHAPTERS_BY_TEXT)
    assert 'cannot convert' in str(exc_info.value)



from typing import Optional

import pyexlatex as pl
from pyexlatex.models.control.documentclass.classtypes.documentclasstype import (
    DocumentClassType,
)

from plufthesis.config import ASSETS_PATH

UF_THESIS_DISSERTATION_CLS_FILE = ASSETS_PATH / "ufdissertation.cls"
UF_THESIS_DISSERTATION_CLS_TEXT = UF_THESIS_DISSERTATION_CLS_FILE.read_text()


def get_class_type(natbib_options: Optional[str] = "numbers") -> DocumentClassType:
    orig_natbib = pl.RequirePackage("natbib", modifier_str="numbers")
    new_natbib = pl.RequirePackage("natbib", modifier_str=natbib_options)
    new_cls_text = UF_THESIS_DISSERTATION_CLS_TEXT.replace(
        str(orig_natbib), str(new_natbib)
    )
    class_type = DocumentClassType("uf-thesis-dissertation", new_cls_text)
    return class_type


# TODO [$602039315315e00008a1abd9]: replace this once pyexlatex has a more convenient way of adding custom class types
def register_doc_type(natbib_options: Optional[str] = "numbers"):
    from pyexlatex.models.control.documentclass.classtypes.custom import (
        CUSTOM_CLASS_TYPES,
    )

    class_type = get_class_type(natbib_options)
    CUSTOM_CLASS_TYPES[class_type.name] = class_type



import datetime
from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Union, List, Optional, Sequence, Callable, Dict

import pyexlatex as pl
from pyexlatex import Package
from pyexlatex.logic.builder import build, _build
from pyexlatex.models.control.documentclass.documentclass import DocumentClass
from pyexlatex.models.section.sections import Chapter
from pyexlatex.models.credits.author import Author
from pyexlatex.models.document import DocumentBase
from pyexlatex.models.format.paragraph.sloppy import Sloppy
from pyexlatex.models.title.title import Title
from pyexlatex.typing import PyexlatexItems
from pyexlatex.models.document import get_table_figure_size_packages
import plufthesis.info_models as im
from plufthesis import register_doc_type

from plufthesis.info_models import ThesisTypes


class UFThesis(DocumentBase):
    """
    The main high-level class for creating UF theses/dissertations
    """
    name = 'document'

    def __init__(self, content: PyexlatexItems, title: str, author: str,
                 major: str, chair: str, abstract: PyexlatexItems,
                 bibliography: pl.Bibliography,
                 dedication_contents: PyexlatexItems = 'I dedicate this to...',
                 acknowledgements_contents: PyexlatexItems = 'I would like to thank...',
                 biographical_contents: PyexlatexItems = 'Nick made this Pyexlatex template from the LaTeX template, then got a Ph.D.',
                 abbreviation_contents: Optional[PyexlatexItems] = None,
                 appendix_contents: Optional[Sequence[Chapter]] = None,
                 multiple_appendices: bool = False,
                 edit_mode: bool = False,
                 bibliography_style: str = 'amsplain',
                 co_chair: Optional[str] = None,
                 thesis_type: ThesisTypes = ThesisTypes.DISSERTATION,
                 degree_type: str = 'Doctor of Philosophy',
                 degree_year: int = datetime.datetime.today().year,
                 degree_month: str = 'May',
                 has_tables: bool = False, has_figures: bool = False,
                 has_objects: bool = False,
                 tables_relative_font_size: int = 0,
                 figures_relative_font_size: int = 0,
                 natbib_options: Optional[str] = 'numbers',
                 packages: Optional[List[Union[Package, str]]]=None,
                 pre_env_contents: Optional[PyexlatexItems] = None,
                 pre_output_func: Optional[Callable] = None):
        self.natbib_options = natbib_options
        self.init_data()
        # These package imports are already handled in the cls file. To avoid conflicts
        # when using items which also include these packages, add these to packages first
        self.data.packages.append(pl.Package('hyperref', modifier_str='linktoc=all'))
        self.data.packages.append(pl.Package('natbib', modifier_str=natbib_options))

        # Necessary to get self.data.references which sets self.has_references in DocumentBase
        # which enables bibtex
        self.data.references.extend(bibliography.references)

        register_doc_type(self.natbib_options)
        if edit_mode:
            self.document_class_obj = DocumentClass(
                document_type='uf-thesis-dissertation',
                options=['editMode'],
            )
        else:
            self.document_class_obj = DocumentClass(
                document_type='uf-thesis-dissertation',
            )
        register_doc_type()

        from pyexlatex.models.item import ItemBase
        if pre_env_contents is None:
            pre_env_contents_list = []
        elif isinstance(pre_env_contents, (ItemBase, str)):
            pre_env_contents_list = [pre_env_contents]
        else:
            pre_env_contents_list = pre_env_contents  # type: ignore

        if has_tables:
            pre_env_contents_list.append(pl.Raw(r'\haveTablestrue'))
        if has_figures:
            pre_env_contents_list.append(pl.Raw(r'\haveFigurestrue'))
        if has_objects:
            pre_env_contents_list.append(pl.Raw(r'\haveObjectstrue'))

        self.temp_tex_contents: Dict[str, PyexlatexItems] = dict(
            dedicationFile=dedication_contents,
            acknowledgementsFile=acknowledgements_contents,
            abstractFile=abstract,
            referenceFile=_build(bibliography.references),
            biographyFile=biographical_contents,
        )
        set_ref_file = r'\setReferenceFile{referenceFile}{' + bibliography_style + '}%'

        pre_env_contents_list.extend([
            Sloppy(),
            Title(title),
            im.DegreeType(degree_type),
            im.Major(major),
            Author(author, short_author=None),
            im.ThesisType(thesis_type),
            im.DegreeYear(degree_year),
            im.DegreeMonth(degree_month),
            im.Chair(chair, co_chair),
            pl.Raw(r"""
\setDedicationFile{dedicationFile}%                 Dedication Page
\setAcknowledgementsFile{acknowledgementsFile}%     Acknowledgements Page
\setAbstractFile{abstractFile}%                     Abstract Page (This should only include the abstract itself)
\setBiographicalFile{biographyFile}%                Biography file of the Author (you).
            """),
            pl.Raw(set_ref_file),
        ])

        if multiple_appendices:
            pre_env_contents_list.append(pl.Raw(r"""
\multipleAppendixtrue%                          Uncomment this if you have more than one appendix, 
%                                                   comment it if you have only one appendix.
            """))

        if abbreviation_contents is not None:
            pre_env_contents_list.append(pl.Raw(r"""
\setAbbreviationsFile{abbreviations}%           Abbreviations Page
            """))
            self.temp_tex_contents.update(abbreviations=abbreviation_contents)

        if appendix_contents is not None:
            pre_env_contents_list.append(pl.Raw(r"""
\setAppendixFile{appendix}%                     Appendix Content; hyperlinking might be weird.
                        """))
            self.temp_tex_contents.update(appendix=appendix_contents)

        for temp_content in self.temp_tex_contents.values():
            self.add_data_from_content(temp_content)

        floatrow_options = 'capposition=top,captionskip=0pt'
        pre_env_contents_list.extend(
            get_table_figure_size_packages(
                tables_relative_font_size,
                figures_relative_font_size,
                floatrow_table_options=floatrow_options,
                floatrow_figure_options=floatrow_options,
            )
        )

        super().__init__(
            content,
            packages=packages,
            pre_env_contents=pre_env_contents_list,
            pre_output_func=pre_output_func,
        )

    def _write_temp_tex_file(self, name: str, content: PyexlatexItems, directory: Path, ext: str = 'tex'):
        out_path = directory / f'{name}.{ext}'
        text_content = build(content)
        out_path.write_text(text_content)
        self.data.filepaths.append(str(out_path))
        self.data.binaries.append(bytes(text_content, 'utf8'))

    def _write_temp_tex_files(self, directory: Path):
        for name, content in self.temp_tex_contents.items():
            if name == 'referenceFile':
                ext = 'bib'
            else:
                ext = 'tex'
            self._write_temp_tex_file(name, content, directory, ext=ext)

    def to_pdf(self, *args, **kwargs):
        with TemporaryDirectory() as tmp:
            tmp_dir = Path(tmp)
            orig_data = deepcopy(self.data)
            self._write_temp_tex_files(tmp_dir)
            result = super().to_pdf(*args, **kwargs)
            self.data = orig_data
        return result

    def to_html(self, *args, **kwargs):
        with TemporaryDirectory() as tmp:
            tmp_dir = Path(tmp)
            orig_data = deepcopy(self.data)
            self._write_temp_tex_files(tmp_dir)
            result = super().to_html(*args, **kwargs)
            self.data = orig_data
        return result



from pathlib import Path

GENERATE_MODE: bool = False

TESTS_PATH = Path(__file__).parent
GENERATED_FILES_DIR = TESTS_PATH / 'generated_files'
INPUT_FILES_DIR = TESTS_PATH / 'input_files'

for path in [GENERATED_FILES_DIR, INPUT_FILES_DIR]:
    if not path.exists():
        path.mkdir()


from copy import deepcopy

import pytest
import pyexlatex as pl
import pandas as pd

from plufthesis.info_models import ThesisTypes
from plufthesis.thesis import UFThesis
from plufthesis.transformers import elevate_sections_by_one_level
from tests.config import INPUT_FILES_DIR

EXAMPLE_BODY = [
    pl.Chapter(
        [
            'Some content',
            pl.UnorderedList(['a', 'b', 'c']),
        ],
        title='First'
    ),
    pl.Chapter(
        [
            'Chapter content',
            pl.Section(
                [
                    'Section content',
                    pl.SubSection(
                        [
                            'Subsection content',
                            pl.SubSubSection(
                                [
                                    'Subsubsubsection content',
                                    pl.CiteP('person'),
                                ],
                                title='First sub sub'
                            )
                        ],
                        title='First sub'
                    )
                ],
                title='First section'
            )
        ],
        title='Second'
    )
]
EXAMPLE_BODY_NEXT_LEVEL_DOWN = [
    pl.Section(
        [
            'Some content',
            pl.UnorderedList(['a', 'b', 'c']),
        ],
        title='First'
    ),
    pl.Section(
        [
            'Chapter content',
            pl.SubSection(
                [
                    'Section content',
                    pl.SubSubSection(
                        [
                            'Subsection content',
                            pl.Paragraph(
                                [
                                    'Subsubsubsection content',
                                    pl.CiteP('person'),
                                ],
                                title='First sub sub'
                            )
                        ],
                        title='First sub'
                    )
                ],
                title='First section'
            )
        ],
        title='Second'
    )
]
TITLE = 'Insert Title Here'
AUTHOR = 'Nick DeRobertis'
MAJOR = 'Insert Major Here'
CHAIR = 'Insert Chair Here'
ABSTRACT = "Insert abstract here"
BIBLIOGRAPHY = pl.Bibliography([pl.BibTexArticle('person', 'Last, First', 'An article title', 'Journal of Journals', '2021', '10', '1', '255-256')])
DEGREE_YEAR = 2021

EXAMPLE_DF = pd.DataFrame(
    [
        (1, 2),
        (3, 4)
    ],
    columns=('a', 'b')
)
GRAPHIC_FILE = INPUT_FILES_DIR / 'nd-logo.png'
TABLES_FIGURES_CHAPTER = pl.Chapter(
    [
        'Some text',
        pl.Table.from_list_of_lists_of_dfs([[EXAMPLE_DF]], caption='My Table'),
        'More text',
        pl.Figure.from_dict_of_names_and_filepaths({'My Figure': str(GRAPHIC_FILE)})
    ],
    title='My Tables and Figures'
)
EXAMPLE_BODY_WITH_TABLES_FIGURES = deepcopy(EXAMPLE_BODY)
EXAMPLE_BODY_WITH_TABLES_FIGURES.append(TABLES_FIGURES_CHAPTER)

# Optional arguments
ABBREVIATIONS = 'ABC is for the alphabet',
APPENDIX_ONE = pl.Chapter('First appendix content', 'Appendix One')
APPENDIX_TWO = pl.Chapter('Second appendix content', 'Appendix Two')
SINGLE_APPENDIX = (APPENDIX_ONE,)
MULTIPLE_APPENDIX = (APPENDIX_ONE, APPENDIX_TWO)
CO_CHAIR = 'Insert Co-Chair Here'
THESIS_TYPE = ThesisTypes.THESIS
DEGREE_TYPE = 'Master of Science'
DEGREE_MONTH = 'August'

ARGS = (
    EXAMPLE_BODY,
    TITLE,
    AUTHOR,
    MAJOR,
    CHAIR,
    ABSTRACT,
    BIBLIOGRAPHY
)

KWARGS = dict(
    degree_year=DEGREE_YEAR
)


@pytest.fixture(scope='session')
def thesis():
    return UFThesis(*ARGS, **KWARGS)


@pytest.fixture(scope='session')
def thesis_full_single_appendix():
    return UFThesis(
        *ARGS, **KWARGS,
        abbreviation_contents=ABBREVIATIONS,
        appendix_contents=SINGLE_APPENDIX,
        co_chair=CO_CHAIR,
        thesis_type=THESIS_TYPE,
        degree_type=DEGREE_TYPE,
        degree_month=DEGREE_MONTH,
    )


@pytest.fixture(scope='session')
def thesis_multiple_appendix():
    return UFThesis(
        *ARGS, **KWARGS,
        appendix_contents=MULTIPLE_APPENDIX,
        multiple_appendices=True,
    )


@pytest.fixture(scope='session')
def thesis_custom_doc_class():
    return UFThesis(
        *ARGS, **KWARGS,
        natbib_options=None,
        bibliography_style='apalike',
    )


@pytest.fixture(scope='session')
def thesis_tables_and_figures():
    tf_args = (
        EXAMPLE_BODY_WITH_TABLES_FIGURES,
        *ARGS[1:]
    )
    return UFThesis(
        *tf_args, **KWARGS,
        has_figures=True,
        has_tables=True,
    )


@pytest.fixture(scope='session')
def thesis_from_next_level_down():
    tf_args = (
        EXAMPLE_BODY_NEXT_LEVEL_DOWN,
        *ARGS[1:]
    )
    return UFThesis(*tf_args, **KWARGS, pre_output_func=elevate_sections_by_one_level)


@pytest.fixture(scope='session')
def thesis_appendix_tables_and_figures():
    return UFThesis(
        *ARGS, **KWARGS,
        has_figures=True,
        has_tables=True,
        appendix_contents=[TABLES_FIGURES_CHAPTER],
        multiple_appendices=True,
    )


@pytest.fixture(scope='session')
def thesis_small_tables_and_figures():
    tf_args = (
        EXAMPLE_BODY_WITH_TABLES_FIGURES,
        *ARGS[1:]
    )
    return UFThesis(
        *tf_args, **KWARGS,
        has_figures=True,
        has_tables=True,
        tables_relative_font_size=-2,
        figures_relative_font_size=-2,
    )


from pyexlatex.models.document import DocumentBase

from tests.config import GENERATED_FILES_DIR, INPUT_FILES_DIR, GENERATE_MODE
from tests.utils.pdf import compare_pdfs


def compare_doc(doc: DocumentBase, name: str):
    if GENERATE_MODE:
        return doc.to_pdf(INPUT_FILES_DIR, outname=name)

    compare_path = INPUT_FILES_DIR / f"{name}.tex"
    assert str(doc) == compare_path.read_text()
    doc.to_pdf(GENERATED_FILES_DIR, outname=name)
    compare_pdfs(INPUT_FILES_DIR / f'{name}.pdf', GENERATED_FILES_DIR / f'{name}.pdf')


from pathlib import Path

import pyexlatex as pl

from tests.config import INPUT_FILES_DIR
from tests.utils.doc import compare_doc
from tests.fixtures.thesis import (
    thesis,
    thesis_full_single_appendix,
    thesis_multiple_appendix,
    thesis_custom_doc_class,
    thesis_tables_and_figures,
    thesis_from_next_level_down,
    thesis_appendix_tables_and_figures,
    thesis_small_tables_and_figures
)


def test_create_thesis(thesis):
    name = "thesis"
    doc = thesis

    compare_doc(doc, name)


def test_create_thesis_full_single_appendix(thesis_full_single_appendix):
    name = "thesis_full_single_appendix"
    doc = thesis_full_single_appendix

    compare_doc(doc, name)


def test_create_thesis_multiple_appendix(thesis_multiple_appendix):
    name = "thesis_multiple_appendix"
    doc = thesis_multiple_appendix

    compare_doc(doc, name)


def test_create_thesis_custom_doc_class(thesis_custom_doc_class):
    name = "thesis_custom_doc_class"
    doc = thesis_custom_doc_class

    compare_doc(doc, name)


def test_create_thesis_tables_and_figures(thesis_tables_and_figures):
    name = "thesis_tables_and_figures"
    doc = thesis_tables_and_figures

    compare_doc(doc, name)


def test_create_thesis_from_next_level_down(thesis_from_next_level_down):
    name = "thesis"
    doc = thesis_from_next_level_down

    compare_doc(doc, name)


def test_create_thesis_appendix_tables_and_figures(thesis_appendix_tables_and_figures):
    name = "thesis_appendix_tables_and_figures"
    doc = thesis_appendix_tables_and_figures

    compare_doc(doc, name)


def test_create_thesis_small_tables_and_figures(thesis_small_tables_and_figures):
    name = "thesis_small_tables_and_figures"
    doc = thesis_small_tables_and_figures

    compare_doc(doc, name)


"""
Functions which transform string content
"""
import re
from typing import Type, Sequence, Dict, cast

import pyexlatex as pl
from pyexlatex.models.section.base import SectionBase

SECTION_CLASSES: Sequence[Type[SectionBase]] = [
    pl.Chapter,
    pl.Section,
    pl.SubSection,
    pl.SubSubSection,
    pl.Paragraph,
    pl.SubParagraph,
]
SECTION_CLASSES_BY_NAME: Dict[str, Type[SectionBase]] = {
    cls.name: cls for cls in SECTION_CLASSES
}
NEXT_LEVEL_UP_CLASSES_BY_NAME: Dict[str, Type[SectionBase]] = {}
for name, cls in SECTION_CLASSES_BY_NAME.items():
    down_cls = cls.next_level_down_class
    if down_cls is not None:
        down_name = down_cls.name
        NEXT_LEVEL_UP_CLASSES_BY_NAME[down_name] = cls


def elevate_sections_by_one_level(content: str) -> str:
    """
    Converts Sections to Chapters, SubSections to Sections,
    SubSubSections to SubSections. Raises error if a Chapter is
    encountered

    Useful to use as pre_output_func in UFThesis if the contents
    were originally set up with Section as the highest level
    rather than Chapter

    :param content:
    :return:
    """
    patterns: Dict[str, str] = {}
    for name, cls in SECTION_CLASSES_BY_NAME.items():
        patterns[name] = r"\\" + name + "{"
        patterns[f"begin_{name}"] = r"\\begin{" + name + "}"
        patterns[f"end_{name}"] = r"\\end{" + name + "}"

    # Adapted from tokenizer example in re docs
    full_regex = "|".join(
        f"(?P<{name}>{pattern})" for name, pattern in patterns.items()
    )
    result = re.sub(full_regex, _replace_with_next_level_up_section, content)
    return result


def _replace_with_next_level_up_section(match: re.Match) -> str:
    kind = cast(str, match.lastgroup)
    name = kind.split("_")[-1]
    cls = SECTION_CLASSES_BY_NAME[name]
    new_cls = NEXT_LEVEL_UP_CLASSES_BY_NAME.get(name, None)
    if new_cls is None:
        raise ValueError(
            f"cannot convert {cls} to next level up as it is "
            f"not a next_level_down_class for any other section type. "
            f"Cannot convert {match.group()}"
        )
    new_name = new_cls.name
    if kind.startswith("begin"):
        return r"\begin{" + new_name + "}"
    if kind.startswith("end"):
        return r"\end{" + new_name + "}"

    # Must be plain match like \section{My Section}
    return "\\" + new_name + "{"
